-------------------------------------------------------------------------------
-- Project  ELE8304 : Circuits intégrés à très grande échelle
-------------------------------------------------------------------------------
-- File     riscv_execute.vhd
-- Authors  Titouan Luard <luardtitouan@gmail.com> 
--          Yann Roberge <yann.roberge@polymtl.ca> 
-- Lab      GRM - Polytechnique Montreal
-- Date     2021-12-04
-------------------------------------------------------------------------------
-- Brief    Execute stage of the pipeline
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;


library work;
use work.riscv_pkg.all;

entity riscv_execute is
  port (
    -- Entrées de l'étage Decode
    i_decode_reg : in decode_reg;

    rs1_data : in std_logic_vector(XLEN-1 downto 0);
    rs2_data : in std_logic_vector(XLEN-1 downto 0);

    -- Sorties
    o_pc_transfer : out std_logic;
    o_pc_target   : out std_logic_vector(XLEN-1 downto 0);

    o_stall : out std_logic;
    o_flush : out std_logic;

    -- Entrées externes au pipeline
    i_rstn : in std_logic;
    i_clk  : in std_logic;

    -- Sorties registrées
    o_execute_reg : out execute_reg);
end entity riscv_execute;

architecture beh of riscv_execute is

  signal alu_src2 : std_logic_vector(XLEN-1 downto 0);

  signal shamt : std_logic_vector(SHAMT_WIDTH-1 downto 0);

  -- Sorties
  signal alu_result : std_logic_vector(XLEN-1 downto 0);
  --signal store_data : std_logic;

  alias opcode is i_decode_reg.opcode;
  alias jump   is i_decode_reg.jump;
  alias branch is i_decode_reg.branch;
  alias imm    is i_decode_reg.imm;
  alias pc     is i_decode_reg.pc;

  constant ALL_ZEROES : std_logic_vector(XLEN-1 downto 0) := (others => '0');

  signal dummy : std_logic; -- doesn't do anything
begin

  -- PC transfer
  -- transfert si jump ou si la condition du BEQ est vraie  
  -- flush la pipeline dans ce même cas
  -- toute la gestion de conflit se fait ici
  -- ie. branch prediction
  pc_transfer:
  process (all) is
  begin
    if (jump = '1') or ((branch = '1') and (alu_result = ALL_ZEROES)) then
      o_flush <= '1';
      o_pc_transfer <= '1';
    else
      o_flush <= '0';
      o_pc_transfer <= '0';
    end if;
  end process pc_transfer;

  -- Résoudre le conflit de données en cas de LW
  o_stall <= '1' when opcode = OPCODE_LW else '0';

  -- Résolution des opérations arithmétiques et logiques
  -- TODO: the whole thing is a big mux depending on OPCODE

  -- Shamt: encoded in IMM for I-TYPE, else encoded in RS2 
  pick_shamt:
  process (opcode, rs2_data) is
  begin
    case opcode is
      when OPCODE_ALU_R_TYPE =>
        shamt <= rs2_data(SHAMT_WIDTH-1 downto 0);
      when OPCODE_ALU_I_TYPE =>
        shamt <= imm(SHAMT_WIDTH-1 downto 0);
      when others =>
        shamt <= (others => '0');
    end case;
  end process pick_shamt;

  alu_feeder:
  process (opcode, rs2_data) is
  begin
    -- Select src2
    case opcode is
      when OPCODE_ALU_R_TYPE | OPCODE_BEQ =>
        alu_src2 <= rs2_data;
      when OPCODE_ALU_I_TYPE =>
        alu_src2 <= imm;
      when others =>
        alu_src2 <= (others => '0');
    end case;
  end process alu_feeder;

  -- Instanciate alu
  alu: entity work.riscv_alu
  port map (
    i_arith   => i_decode_reg.arith,
    i_sign    => i_decode_reg.sign,
    i_opcode  => i_decode_reg.alu_opcode,
    i_shamt   => shamt,
    i_src1    => rs1_data,
    i_src2    => alu_src2,
    o_res     => alu_result
  );

  -- Ajustement du PC en cas de saut ou branchement
  -- Instanciate adder
  dut: entity work.riscv_adder
  generic map (
    N => XLEN
  )
  port map (
    i_a    => imm,
    i_b    => pc,
    i_sign => SIGN_UNSIGNED,
    i_sub  => ARITH_ADD,
    o_sum(XLEN) => dummy,
    o_sum(XLEN-1 downto 0) => o_pc_target
  );


  -- Écriture du registre execute
  -- same as decode register write thingy without flush input
  write_reg_ex_me:
  process (i_clk) is
  begin
    if rising_edge(i_clk) then
      if i_rstn = '1' then
        o_execute_reg.opcode <= opcode;

        o_execute_reg.alu_result <= alu_result;
        o_execute_reg.store_data <= rs2_data;
        o_execute_reg.rd_addr    <= i_decode_reg.rd_addr;
      else
        o_execute_reg.opcode <= (others => '0');

        o_execute_reg.alu_result <= (others => '0');
        o_execute_reg.store_data <= (others => '0');
        o_execute_reg.rd_addr    <= (others => '0');
      end if;
    end if;
  end process write_reg_ex_me;

  -- DEBUG
  --o_pc_target <= (others => '1');
  --o_execute_reg.alu_result <= (others => '1');
  --o_execute_reg.store_data <= '1';

end architecture beh;
